
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#ifndef _MOOF_OCTREE_HH_
#define _MOOF_OCTREE_HH_

#include <algorithm>
#include <list>

#include <boost/shared_ptr.hpp>

#include <stlplus/ntree.hpp>

#include <Moof/Aabb.hh>
#include <Moof/Drawable.hh>
#include <Moof/Entity.hh>
#include <Moof/Math.hh>
#include <Moof/Sphere.hh>
	

namespace Mf {


class Camera;


struct OctreeNode : public Entity
{
	std::list<EntityPtr> objects;

	OctreeNode()
	{
		aabb_.min = Vector3(-1.0, -1.0, -1.0);
		aabb_.max = Vector3(1.0, 1.0, 1.0);
		sphere_.init(Vector3(0.0, 0.0, 0.0), 1.41421);
	}

	OctreeNode(const Aabb& aabb)
	{
		aabb_ = aabb;
		sphere_.point = aabb.getCenter();
		sphere_.radius = (aabb.min - sphere_.point).length();
	}

	void draw(Scalar alpha) const
	{
		std::list<EntityPtr>::const_iterator it;

		for (it = objects.begin(); it != objects.end(); ++it)
		{
			(*it)->draw(alpha);
		}

		//if (!objects.empty())
			//aabb_.draw(); // temporary
	}

	void drawIfVisible(Scalar alpha, const Camera& cam) const
	{
		std::list<EntityPtr>::const_iterator it;

		for (it = objects.begin(); it != objects.end(); ++it)
		{
			(*it)->drawIfVisible(alpha, cam);
		}

		//if (!objects.empty())
			//aabb_.draw();
	}


	bool isVisible(const Camera& cam) const
	{
		if (sphere_.isVisible(cam))
		{
			return aabb_.isVisible(cam);
		}

		return false;
	}


	static bool compareZOrder(EntityPtr a, EntityPtr b)
	{
		return a->getSphere().point[2] < b->getSphere().point[2];
	}

	void sort()
	{
		//std::sort(objects.begin(), objects.end(), compareZOrder);
		objects.sort(compareZOrder);
	}
};


class Octree
{
public:

	explicit Octree(const OctreeNode& rootNode)
	{
		tree_.insert(rootNode);
	}

	stlplus::ntree<OctreeNode>::iterator insert(EntityPtr entity)
	{
		return insert(tree_.root(), entity);
	}

	stlplus::ntree<OctreeNode>::iterator reinsert(EntityPtr entity,
			stlplus::ntree<OctreeNode>::iterator node);

	void drawIfVisible(Scalar alpha, const Camera& cam)
	{
		drawIfVisible(tree_.root(), alpha, cam);
	}

	void sort();

private:
	stlplus::ntree<OctreeNode>::iterator insert(stlplus::ntree<OctreeNode>::iterator node, EntityPtr entity);
	
	void addChild(stlplus::ntree<OctreeNode>::iterator node, int index);

	void draw(stlplus::ntree<OctreeNode>::iterator node, Scalar alpha);
	void drawIfVisible(stlplus::ntree<OctreeNode>::iterator node,
			Scalar alpha, const Camera& cam);

	stlplus::ntree<OctreeNode> tree_;
};

typedef boost::shared_ptr<Octree> OctreePtr;


} // namespace Mf

#endif // _MOOF_OCTREE_HH_

/** vim: set ts=4 sw=4 tw=80: *************************************************/

