
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include <iostream>
#include <map>
#include <vector>

#include "Aabb.hh"
#include "Camera.hh"
#include "Deserializer.hh"
#include "Entity.hh"
#include "Math.hh"
#include "Mippleton.hh"
#include "OpenGL.hh"
#include "Scene.hh"
#include "Serializable.hh"
#include "Tilemap.hh"


namespace Mf {


class Scene::Impl : public Mippleton<Impl>
{
	class Quad : public Entity
	{
	public:
		Quad(const Vector3 vertices[4], const std::string& texture,
				Tilemap::Index tileIndex) :
			tilemap_(texture),
			detail_(0),
			blending_(false),
			fog_(false)
		{
			for (int i = 0, num = 0; i < 4; ++i)
			{
				for (int j = 0; j < 3; ++j, ++num)
				{
					vertices_[num] = vertices[i][j];
				}
			}

			if (!tilemap_.getTileCoords(tileIndex, texCoords_))
			{
				std::cerr << "no coords for tile's texture" << std::endl;
			}

			aabb_.encloseVertices(vertices, 4);
			sphere_.point = aabb_.getCenter();
			sphere_.radius = (aabb_.min - sphere_.point).length();
		}

		void setDetail(long detail)
		{
			detail_ = detail;
		}

		void setBlending(bool blending)
		{
			blending_ = blending;
		}

		void setFog(bool fog)
		{
			fog_ = fog;
		}

		void draw(Scalar alpha = 0.0) const
		{
			if (blending_)
			{
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			}

			if (fog_)
			{
				glEnable(GL_FOG);
				glFogi(GL_FOG_MODE, GL_LINEAR);
			}

			glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
			tilemap_.bind();

			glVertexPointer(3, GL_SCALAR, 0, vertices_);
			glTexCoordPointer(2, GL_SCALAR, 0, texCoords_);

			glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

			glDisable(GL_BLEND);
			glDisable(GL_FOG);
		}

		bool isVisible(const Camera& cam) const
		{
			return sphere_.isVisible(cam);
		}

	private:
		Scalar		vertices_[12];
		Scalar		texCoords_[8];
		
		Tilemap		tilemap_;

		long		detail_;
		bool		blending_;
		bool		fog_;
	};


	static void loadBox(Aabb& theBox, SerializablePtr obj)
	{
		std::vector<SerializablePtr> numbers;

		if (obj->get(numbers) && numbers.size() == 6)
		{
			double num;

			if (numbers[0]->getNumber(num)) theBox.min[0] = Scalar(num);
			if (numbers[1]->getNumber(num)) theBox.min[1] = Scalar(num);
			if (numbers[2]->getNumber(num)) theBox.min[2] = Scalar(num);
			if (numbers[3]->getNumber(num)) theBox.max[0] = Scalar(num);
			if (numbers[4]->getNumber(num)) theBox.max[1] = Scalar(num);
			if (numbers[5]->getNumber(num)) theBox.max[2] = Scalar(num);
		}
	}

public:
	Impl(const std::string& name) :
		Mippleton<Impl>(name)
	{
		loadFromFile();
	}


	void loadInstructions(SerializablePtr root)
	{
		std::vector<SerializablePtr> rootObj;
		std::vector<SerializablePtr>::iterator it;

		if (!root->get(rootObj))
		{
			std::cerr << "error loading scene instructions" << std::endl;
			return;
		}

		Matrix4		transform;
		std::string	texture;

		for (it = rootObj.begin(); it != rootObj.end(); ++it)
		{
			std::string instruction;

			if ((*it)->get(instruction))
			{
				if (instruction == "reset_transform")
				{
					transform.identity();
				}
				else if (instruction == "translate")
				{
					std::vector<SerializablePtr> values;

					++it;
					if ((*it)->get(values))
					{
						Vector3 vec;

						for (size_t i = 0; i < values.size(); ++i)
						{
							double value;

							if (values[i]->getNumber(value))
							{
								vec[i] = value;
							}
						}

						Matrix4 translation;
						cml::matrix_translation(translation, vec);
						transform = translation * transform;
					}
				}
				else if (instruction == "scale")
				{
					std::vector<SerializablePtr> values;

					++it;
					if ((*it)->get(values))
					{
						if (values.size() == 1)
						{
							double value = 1.0;

							values[0]->getNumber(value);

							Matrix4 scaling;
							cml::matrix_uniform_scale(scaling,
									Scalar(value));
							transform = scaling * transform;
						}
						else if (values.size() == 3)
						{
							Vector3 vec;

							for (size_t i = 0; i < values.size(); ++i)
							{
								double value;

								if (values[i]->getNumber(value))
								{
									vec[i] = value;
								}
							}

							Matrix4 scaling;
							cml::matrix_scale(scaling, vec);
							transform = scaling * transform;
						}
					}
				}
				else if (instruction == "rotate")
				{
					std::vector<SerializablePtr> values;

					++it;
					if ((*it)->get(values))
					{
						if (values.size() == 2)
						{
							std::string axis;
							size_t index = 0;
							double value = 0.0;

							if (values[0]->get(axis))
							{
								if (axis == "x")      index = 0;
								else if (axis == "y") index = 1;
								else if (axis == "z") index = 2;

								values[1]->getNumber(value);
							}

							cml::matrix_rotate_about_world_axis(transform,
									index, cml::rad(Scalar(value)));
						}
					}
				}
				else if (instruction == "texture")
				{
					++it;
					(*it)->get(texture);
				}
				else if (instruction == "tilemap")
				{
					++it;
					loadTilemap(*it, transform, texture);
				}
				else if (instruction == "billboard")
				{
					++it;
					loadBillboard(*it, transform, texture);
				}
			}
		}
	}


	void loadTilemap(SerializablePtr root, const Matrix4& transform,
			const std::string& texture)
	{
		std::map<std::string,SerializablePtr> rootObj;
		std::map<std::string,SerializablePtr>::iterator it;

		if (!root->get(rootObj))
		{
			std::cerr << "error loading scene tilemap object" << std::endl;
			return;
		}

		long width = 1;
		long height = 1;
		std::vector< std::vector<Tilemap::Index> > indices;

		if ((it = rootObj.find("width")) != rootObj.end())
		{
			(*it).second->get(width);
		}
		else
		{
			std::cerr << "width is a required field of a tilemap" << std::endl;
			return;
		}

		std::vector<SerializablePtr> tiles;

		if ((it = rootObj.find("tiles")) != rootObj.end() &&
				(*it).second->get(tiles) &&
				tiles.size() % width == 0)
		{
			std::vector<SerializablePtr>::iterator jt;
			int w, h;

			height = tiles.size() / width;
			indices.resize(height);

			// the indices are stored upside-down in the scene file so that they
			// are easier to edit as text, so we'll need to load them last row
			// first

			for (h = height - 1, jt = tiles.begin(); jt != tiles.end(); --h)
			{
				std::vector<Tilemap::Index> row;

				for (w = 0; w < width && jt != tiles.end(); ++w, ++jt)
				{
					long index;

					if ((*jt)->get(index))
					{
						row.push_back(Tilemap::Index(index));
					}
				}

				indices[h] = row;
			}
		}
		else
		{
			std::cerr << "error loading tiles from tilemap object" << std::endl;
			return;
		}

		Vector4 vertices[height+1][width+1];

		Matrix4 transposedTransform = transform;
		transposedTransform.transpose();

		for (int h = 0; h <= height; ++h)
		{
			for (int w = 0; w <= width; ++w)
			{
				vertices[h][w] = Vector4(Scalar(w), Scalar(h), 0.0, 1.0) *
					transposedTransform;
			}
		}

		for (int h = 0; h < height; ++h)
		{
			for (int w = 0; w < width; ++w)
			{
				if (indices[h][w] == Tilemap::NO_TILE) continue;

				Vector3 quadVertices[4];

				demoteVector(quadVertices[0], vertices[h][w]);
				demoteVector(quadVertices[1], vertices[h][w+1]);
				demoteVector(quadVertices[2], vertices[h+1][w+1]);
				demoteVector(quadVertices[3], vertices[h+1][w]);

				Quad* quad = new Quad(quadVertices, texture, indices[h][w]);
				boost::shared_ptr<Quad> quadPtr(quad);

				octree->insert(quadPtr);
			}
		}
	}

	void loadBillboard(SerializablePtr root, const Matrix4& transform,
			const std::string& texture)
	{
		std::map<std::string,SerializablePtr> rootObj;
		std::map<std::string,SerializablePtr>::iterator it;

		Tilemap::Index	index = 0;
		long			width = 1;
		bool			blending = false;
		bool			fog = false;

		if (root->get(rootObj))
		{
			if ((it = rootObj.find("tile")) != rootObj.end())
			{
				long value;
				if ((*it).second->get(value))
				{
					index = Tilemap::Index(value);
				}
			}

			if ((it = rootObj.find("u_scale")) != rootObj.end())
			{
				(*it).second->get(width);
			}

			if ((it = rootObj.find("blend")) != rootObj.end())
			{
				(*it).second->get(blending);
			}

			if ((it = rootObj.find("fog")) != rootObj.end())
			{
				(*it).second->get(fog);
			}
		}


		Vector4 vertices[2][width+1];

		Matrix4 transposedTransform = transform;
		transposedTransform.transpose();

		Scalar xf;
		Scalar increment = 1.0 / Scalar(width);

		for (int h = 0; h <= 1; ++h)
		{
			xf = 0.0;
			for (int w = 0; w <= width; ++w, xf += increment)
			{
				vertices[h][w] = Vector4(xf, Scalar(h), 0.0, 1.0) *
					transposedTransform;
			}
		}

		for (int w = 0; w < width; ++w)
		{
			Vector3 quadVertices[4];

			demoteVector(quadVertices[0], vertices[0][w]);
			demoteVector(quadVertices[1], vertices[0][w+1]);
			demoteVector(quadVertices[2], vertices[1][w+1]);
			demoteVector(quadVertices[3], vertices[1][w]);

			Quad* quad = new Quad(quadVertices, texture, index);
			quad->setBlending(blending);
			quad->setFog(fog);

			boost::shared_ptr<Quad> quadPtr(quad);

			octree->insert(quadPtr);
		}
	}


	void loadFromFile()
	{
		std::string filePath = Scene::getPath(getName());

		Deserializer	deserializer(filePath, true);
		SerializablePtr	root = deserializer.deserialize();

		std::map<std::string,SerializablePtr> rootObj;
		std::map<std::string,SerializablePtr>::iterator it;

		if (!root || !root->get(rootObj))
		{
			std::cerr << "error loading scene file" << std::endl;
			return;
		}

		if ((it = rootObj.find("playfield_bounds")) != rootObj.end())
		{
			loadBox(playfieldBounds, (*it).second);
		}
		if ((it = rootObj.find("maximum_bounds")) != rootObj.end())
		{
			loadBox(maximumBounds, (*it).second);
		}
		else
		{
			std::cerr << "maximum bounds required in scene" << std::endl;
			return;
		}

		// create the tree to store the quads
		octree = OctreePtr(new Octree(maximumBounds));

		if ((it = rootObj.find("instructions")) != rootObj.end())
		{
			loadInstructions((*it).second);
		}

		octree->sort();
	}


	void draw(Scalar alpha, const Camera& cam) const
	{
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		octree->drawIfVisible(alpha, cam);

		//glDisableClientState(GL_VERTEX_ARRAY);
		//glDisableClientState(GL_TEXTURE_COORD_ARRAY);

		//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

		//Texture::resetBind();
		//glColor3f(0.0f, 1.0f, 0.0f);
		//playfieldBounds.draw();
		//glColor3f(0.0f, 0.0f, 1.0f);
		//maximumBounds.draw();

		//glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}


	Aabb playfieldBounds;
	Aabb maximumBounds;

	OctreePtr octree;
};


Scene::Scene(const std::string& name) :
	// pass through
	impl_(Scene::Impl::retain(name), &Scene::Impl::release) {}


void Scene::draw(Scalar alpha, const Camera& cam) const
{
	// pass through
	impl_->draw(alpha, cam);
}

void Scene::refresh()
{
	//impl_->objects.clear();
	impl_->loadFromFile();
}


OctreePtr Scene::getOctree() const
{
	// pass through
	return impl_->octree;
}

/**
 * Specialized search location for scene files.  They can be found in the
 * "scenes" subdirectory of any of the searched directories.
 */

std::string Scene::getPath(const std::string& name)
{
	return Resource::getPath("scenes/" + name + ".json");
}


} // namespace Mf

/** vim: set ts=4 sw=4 tw=80: *************************************************/

