
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include <iostream>
#include <string>

#include <SDL/SDL.h>
#include <SDL/SDL_sound.h>
#include <AL/al.h>

#include "Mippleton.hh"
#include "Sound.hh"

#define BUFFER_SIZE (8 * 4096)

namespace Mf {


struct Sound::Impl
{

	static ALenum getAudioFormat(const Sound_AudioInfo& audioInfo)
	{
		if (audioInfo.format == AUDIO_U8 || audioInfo.format == AUDIO_S8)
		{
			if (audioInfo.channels == 1) return AL_FORMAT_MONO8;
			else                         return AL_FORMAT_STEREO8;
		}
		else
		{
			if (audioInfo.channels == 1) return AL_FORMAT_MONO16;
			else                         return AL_FORMAT_STEREO16;
		}
	}
	
	struct Buffer : public Mippleton<Buffer>
	{
		Buffer(const std::string& name) :
			Mippleton<Buffer>(name)
		{
			objects[0] = 0;
			objects[1] = 0;
		}

		~Buffer()
		{
			alDeleteBuffers(2, objects);

			if (sound) Sound_FreeSample(sound);
		}


		void loadFromFile(const std::string& filePath, bool stream)
		{
			if (objects[0] != 0) return;

			sound = Sound_NewSampleFromFile(filePath.c_str(),
					NULL, BUFFER_SIZE);

			if (!sound)
			{
				std::cerr << "could not load sound from file" << std::endl;
				exit(1);
			}

			if (!stream)
			{
			unsigned decoded = Sound_DecodeAll(sound);
			if (decoded == 0)
			{
				std::cout << "decoded no bytes" << std::endl;
				exit(1);
			}

			alGenBuffers(2, objects);
			alBufferData(objects[0], getAudioFormat(sound->actual), sound->buffer,
					sound->buffer_size, sound->actual.rate);
			std::cerr << "buffer size: " << sound->buffer_size << std::endl;
			std::cerr << "channels: " << (int)sound->actual.channels << std::endl;
			std::cerr << "format: " << sound->actual.format << std::endl;
			std::cerr << "frequency: " << sound->actual.rate << std::endl;

			Sound_FreeSample(sound);
			sound = 0;
			}
			else
			{
			std::cerr << "buffer size: " << sound->buffer_size << std::endl;
			std::cerr << "channels: " << (int)sound->actual.channels << std::endl;
			std::cerr << "format: " << sound->actual.format << std::endl;
			std::cerr << "frequency: " << sound->actual.rate << std::endl;
				alGenBuffers(2, objects);
			}
		}

		bool stream(ALuint buffer)
		{
			int bytes = Sound_Decode(sound);

			if (bytes < BUFFER_SIZE) return false;

			alBufferData(buffer, getAudioFormat(sound->actual), sound->buffer,
					sound->buffer_size, sound->actual.rate);
			return false;
		}

		Sound_Sample* sound;
		ALuint objects[2];

	//ALfloat location[] = {0.0f, 0.0f, 0.0f};
	//ALfloat location2[] = {0.0f, 0.0f, 0.0f};
	//ALfloat orient[] = {0.0f, 0.0f, -1.0f, 0.0, 1.0, 0.0};


	//alListenerfv(AL_POSITION, location);
	//alListenerfv(AL_VELOCITY, location);
	//alListenerfv(AL_VELOCITY, orient);
	};

	Impl(const std::string& name, bool stream = false) :
		buffer_(Buffer::retain(name), Buffer::release)
	{
		if (!stream) buffer_->loadFromFile(Sound::getPath(name), stream);
		else         buffer_->loadFromFile(SoundStream::getPath(name), stream);

		ALfloat location[] = {0.0f, 0.0f, 0.0f};
		
		alGenSources(1, &source_);

		alSourcef(source_,  AL_PITCH, 1.0f);
		alSourcef(source_,  AL_GAIN, 1.0f);
		alSourcefv(source_, AL_POSITION, location);
		alSourcefv(source_, AL_VELOCITY, location);
		alSourcei(source_,  AL_LOOPING, AL_FALSE);

		if (!stream)
		{
			alSourcei(source_, AL_BUFFER, buffer_->objects[0]);
		}
		else
		{
			buffer_->stream(buffer_->objects[0]);
			buffer_->stream(buffer_->objects[1]);

			alSourceQueueBuffers(source_, 2, buffer_->objects);
		}
	}

	~Impl()
	{
		alDeleteSources(1, &source_);
	}


	void update()
	{
		int finished = 0;

		alGetSourcei(source_, AL_BUFFERS_PROCESSED, &finished);

		while (finished-- > 0)
		{
			ALuint buffer;

			alSourceUnqueueBuffers(source_, 1, &buffer);
			buffer_->stream(buffer);
			alSourceQueueBuffers(source_, 1, &buffer);
		}
	}


	boost::shared_ptr<Buffer>	buffer_;
	ALuint						source_;
	bool playing;
};


Sound::Sound(const std::string& name) :
	// pass through
	impl_(new Sound::Impl(name)) {}


void Sound::play()
{
	//alSourceRewind(impl_->source_);
	alSourcePlay(impl_->source_);
	impl_->playing = true;
}

void Sound::pause()
{
	alSourcePause(impl_->source_);
	impl_->playing = false;
}

void Sound::togglePlayPause()
{
	if (impl_->playing) pause();
	else play();
}

void Sound::setGain(Scalar gain)
{
	alSourcef(impl_->source_, AL_GAIN, gain);
}


std::string Sound::getPath(const std::string& name)
{
	std::string path = Resource::getPath("sounds/" + name + ".ogg");
	return path;
}


//##############################################################################


SoundStream::SoundStream(const std::string& name)
	// pass through
	//impl_(name, true) {}
{
	impl_ = boost::shared_ptr<Sound::Impl>(new Sound::Impl(name, true));
}


void SoundStream::update(Scalar t, Scalar dt)
{
	// pass through
	impl_->update();
}


std::string SoundStream::getPath(const std::string& name)
{
	std::string path = Resource::getPath("sounds/" + name + ".xm");
	return path;
}


} // namespace Mf

/** vim: set ts=4 sw=4 tw=80: *************************************************/

